package com.ifgi.gnc.checker;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.logging.Logger;

import com.ifgi.gnc.client.BBox;
import com.ifgi.gnc.client.Product;
import com.ifgi.gnc.constants.GODConstants;
import com.ifgi.gnc.constants.Products;
import com.ifgi.gnc.server.Emailer;
import com.ifgi.gnc.server.ProductInstance;
import com.ifgi.gnc.server.writer.ScriptWriter;

/**
 * A thread that checks the gnc database for jobs that shall be processed. This can be delete, script writing or execution operations.
 * 
 * @author Christoph Mlligann
 *
 */

public class Checker extends Thread {

	private GODConstants constants = new GODConstants();
	private Products products = new Products();
	private Logger logger = Logger.getLogger("Checker_log");

	private final Timestamp retrievalBuffer = new Timestamp(300000); // delay with which data is requested
	private final Timestamp dropBuffer = new Timestamp(604800000); // time after that a job is finally cancelled
	private final Timestamp updateRate = new Timestamp(300000); // frequency of table updates

	/**
	 * Start the checker. It will never stop by itself.
	 */
	public void run() {
		while (!this.isInterrupted()) {
			long before = System.currentTimeMillis();
			createScriptsFromDB();
			processDBJobsUntil(new Timestamp(System.currentTimeMillis() - retrievalBuffer.getTime()));
			deleteDBJobsBefore(new Timestamp(System.currentTimeMillis() - dropBuffer.getTime() - retrievalBuffer.getTime()));
			long after = System.currentTimeMillis();
			long diff =  updateRate.getTime() - (after - before);
			// sleep the rest of the time that remains from updateRate
			if (diff > 0) {
				try {
					Thread.sleep(diff);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	/**
	 * Write and deploy script and info files for all activated jobs where it is still missing.
	 * @return true if successful, false if there have been database or file errors
	 */
	private boolean createScriptsFromDB() {
		Connection conn;
		try {
			conn = getDBConnection();
		} catch (SQLException e) {
			logger.severe("Unable to create database connection: SQL error code " + e.getErrorCode());
			return false;
		}
		Statement stm = null;
		try {
			stm = conn.createStatement();
			ResultSet rs = stm.executeQuery("SELECT * FROM " + constants.DB_TABLE + " WHERE " + constants.DB_COL_SCRIPT + " IS NOT NULL");
			while (rs.next()) {
				String scriptFile = rs.getString(constants.DB_COL_SCRIPT);
				if (!isFileExisting(constants.SCRIPT_DIR + scriptFile + ".bat")) {
					Product product = products.byName(rs.getString(constants.DB_COL_PRODUCT));
					double topRightLat = rs.getDouble(constants.DB_COL_TOPRIGHTLAT);
					double topRightLon = rs.getDouble(constants.DB_COL_TOPRIGHTLON);
					double bottomLeftLat = rs.getDouble(constants.DB_COL_BOTTOMLEFTLAT);
					double bottomLeftLon = rs.getDouble(constants.DB_COL_BOTTOMLEFTLON);
					BBox bbox = new BBox(topRightLat, topRightLon, bottomLeftLat, bottomLeftLon);
					Timestamp ts = rs.getTimestamp(constants.DB_COL_EXECUTION);
					String format = rs.getString(constants.DB_COL_FORMAT);
					ProductInstance pi = new ProductInstance(product, bbox, ts, format);
					ScriptWriter sw = ScriptWriter.getScriptWriter(pi.getMetaData().getName());
					String email = rs.getString(constants.DB_COL_EMAIL);
					String destinationFolder = generateFolderByEmail(email);
					try {
						sw.createScript(pi, scriptFile + ".bat", constants.SCRIPT_DIR, destinationFolder);
					} catch (IOException e) {
						logger.severe("Unable to create script file " + scriptFile);
						return false;
					} catch (InterruptedException e) {
						logger.severe("Process execution interrupted while creating script file " + scriptFile);
						return false;
					}
					if (!pi.write2File(destinationFolder + scriptFile + ".xml")) {
						logger.warning("Error writing product instance information for " + scriptFile + "." + pi.getFormat());
					}
				}
			}
		} catch (SQLException e) {
			logger.warning("Unable to execute delete: SQL error code " + e.getErrorCode());
			return false;
		}
		try {
			conn.close();
		} catch (SQLException e) {
			logger.warning("Unable to close database connection.");
			return false;
		}
		return true;
	}

	/**
	 * Delete jobs and everything that comes along with it before a certain timestamp.
	 * @param timestamp the time before that jobs are deleted
	 * @return true if successful, false if there have been database errors
	 */
	private boolean deleteDBJobsBefore(Timestamp timestamp) {
		Connection conn;
		try {

			conn = getDBConnection();
		} catch (SQLException e) {
			logger.severe("Unable to create database connection: SQL error code " + e.getErrorCode());
			return false;
		}
		Statement stm = null;
		try {
			stm = conn.createStatement();
			ResultSet rs = stm.executeQuery("SELECT * FROM " + constants.DB_TABLE + " WHERE " + constants.DB_COL_EXECUTION + " <= '" + timestamp + "' AND " + constants.DB_COL_SCRIPT + " IS NOT NULL");
			while (rs.next()) {
				logIrretrievable(rs.getString(constants.DB_COL_EMAIL), rs.getString(constants.DB_COL_PRODUCT), rs.getTimestamp(constants.DB_COL_EXECUTION));
				String destinationFolder = generateFolderByEmail(rs.getString(constants.DB_COL_EMAIL));
				String scriptFile = constants.SCRIPT_DIR + rs.getString(constants.DB_COL_SCRIPT) + ".bat";
				String infoFile = destinationFolder + rs.getString(constants.DB_COL_SCRIPT) + ".xml";
				removeFile(scriptFile);
				removeFile(infoFile);
			}
			String delSt = "DELETE FROM " + constants.DB_TABLE + " WHERE " + constants.DB_COL_EXECUTION + " <= '" + timestamp + "' AND " + constants.DB_COL_SCRIPT + " IS NOT NULL";
			stm.executeUpdate(delSt);
		} catch (SQLException e) {
			logger.warning("Unable to execute delete: SQL error code " + e.getErrorCode());
			return false;
		}
		try {
			conn.close();
		} catch (SQLException e) {
			logger.warning("Unable to close database connection.");
			return false;
		}
		return true;
	}
	
	/**
	 * Executes scripts and checks whether they have been successful. 
	 * @param timestamp time before that jobs are processed
	 * @return true if successful, false if there have been database errors
	 */
	private boolean processDBJobsUntil(Timestamp timestamp) {
		Connection conn;
		try {
			conn = getDBConnection();
		} catch (SQLException e) {
			logger.severe("Unable to create database connection: SQL error code " + e.getErrorCode());
			return false;
		}
		Statement stm = null;
		ArrayList<String> toDelete = new ArrayList<String>();
		try {
			stm = conn.createStatement();
			ResultSet rs = stm.executeQuery("SELECT * FROM " + constants.DB_TABLE + " WHERE " + constants.DB_COL_EXECUTION + " <= '" + timestamp + "' AND " + constants.DB_COL_SCRIPT + " IS NOT NULL");
			while (rs.next()) {
				String destinationFolder = generateFolderByEmail(rs.getString(constants.DB_COL_EMAIL));
				String scriptFile = constants.SCRIPT_DIR + rs.getString(constants.DB_COL_SCRIPT) + ".bat";
				String imageFile = destinationFolder + rs.getString(constants.DB_COL_SCRIPT) + "." + rs.getString(constants.DB_COL_FORMAT);
				if (isFileExisting(imageFile) || executeScript(scriptFile)) {
					if (isFileExisting(imageFile)) {
						toDelete.add(rs.getString(constants.DB_COL_SCRIPT));
						logRetrievable(	rs.getString(constants.DB_COL_EMAIL),
											rs.getString(constants.DB_COL_PRODUCT),
											rs.getTimestamp(constants.DB_COL_EXECUTION),
											rs.getString(constants.DB_COL_SCRIPT) + "." + rs.getString(constants.DB_COL_FORMAT));
						removeFile(scriptFile);
					} else {
						logger.warning("Image " + imageFile + " not (yet) available. Job left pending.");
					}
				} else {
					logger.warning("Script " + scriptFile + " not (yet) available. Job left pending.");
				}
			}
		} catch (SQLException e) {
			logger.warning("Unable to execute query: SQL error code " + e.getErrorCode());
			return false;
		}	
		for (String script : toDelete) {
			try {
				stm = conn.createStatement();
				String delSt = "DELETE FROM " + constants.DB_TABLE + " WHERE " + constants.DB_COL_SCRIPT + " = '" + script + "'";
				stm.executeUpdate(delSt);
			} catch (SQLException e) {
				logger.warning("Unable to execute delete: SQL error code " + e.getErrorCode());
				return false;
			}
		}
		try {
			conn.close();
		} catch (SQLException e) {
			logger.warning("Unable to close database connection.");
			return false;
		}
		return true;
	}
	
	/**
	 * Generates a folder path by id and even sets up the physical user folder if it is not yet there.
	 * @param eMail the email, or any other unique identifier
	 * @return the folder's path
	 */
	private String generateFolderByEmail(String eMail) {
		String name = eMailEncode(eMail);
		String path = constants.USER_DIR + name + "\\";
		String url = constants.USER_URL + name;
		if (!isFileExisting(path)) {
			createUserFolder(path);
			createLogFile(path, name);
			sendInitialMail(eMail, url);
		}
		return path;
	}
	
	/**
	 * Initializes a log file.
	 * @param path the file's path
	 * @param name the filename
	 */
	private void createLogFile(String path, String name) {
		File logFile = new File(path + name + ".txt");
		try {
			PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(logFile)));
			pw.println(new Timestamp(System.currentTimeMillis()) + " - log file created in folder " + name);
			pw.close();
		} catch (IOException e) {
			logger.warning("Unable to write to log file " + logFile.getAbsolutePath());
		}
		
	}
	
	/**
	 * Just creates a folder.
	 * @param user a path
	 */
	private void createUserFolder(String user) {
		new File(user).mkdir();
	}

	/**
	 * Sends a mail with all the information the gnc needs.
	 * @param eMail the eMail address
	 * @param url the website the user is pointed to
	 */
	private void sendInitialMail(String eMail, String url) {
		String separator = System.getProperty("line.separator");
		String subject = "GEONETCast On Demand - Your Data Folder";
		String message = 	"Dear GOD User," + separator + separator + 
							"GOD server has created a folder for your scheduled data. You find it on the following URL: " + separator + separator + 
							url + separator + separator +
							"Keep the mail for personal reference only. Access to the data is possible with vpn connection to the university network." + separator + 
							"Image files will be retrievable as soon as possible. In case of errors and additional information have a look at the log file: " + separator + 
							url + eMailEncode(eMail) + ".txt" + separator + separator +
							"Your -GEONETCast On Demand- Team.";
		new Emailer(eMail, subject, message);
	}
	
	/**	
	 * @return the connection
	 * @throws SQLException failed user authenticification or missing db driver 
	 */
	private Connection getDBConnection() throws SQLException {
		try {
			Class.forName(constants.DB_DRIVER);
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		Connection conn = null;
		String url = "jdbc:postgresql:" + (constants.DB_HOST != null ? ("//" + constants.DB_HOST) + (constants.DB_PORT != null ? ":" + constants.DB_PORT : "") + "/" : "") + constants.DB;
		try {
			conn = DriverManager.getConnection(url, constants.DB_USER, constants.DB_PW);
		} catch (SQLException e) {
			throw e;
		}
		return conn;
	}
	
	/**
	 * Writes a log entry for a deleted job.
	 * @param email the user id
	 * @param product a product identifier
	 * @param execution missed time
	 */
	private void logIrretrievable(String email, String product, Timestamp execution) {
		String logPath = constants.USER_DIR + eMailEncode(email) + "\\" + eMailEncode(email) + ".txt"; 
		File logFile = new File(logPath);
		try {
			PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(logFile, true)), true);
			pw.println(new Timestamp(System.currentTimeMillis()) + " - Job cancelled: Product " + product + ", scheduled for " + execution);
			pw.close();
		} catch (IOException e) {
			logger.warning("Unable to write to log file " + logFile.getAbsolutePath());
		}
	}
	
	/**
	 * Writes a log entry for a successful job.
	 * @param email the user id
	 * @param product a product identifier
	 * @param execution execution time
	 * @param filename where you can find it
	 */
	private void logRetrievable(String email, String product, Timestamp execution, String filename) {
		String logPath = constants.USER_DIR + eMailEncode(email) + "\\" + eMailEncode(email) + ".txt";
		File logFile = new File(logPath);
		try {
			PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(logFile, true)), true);
			pw.println(new Timestamp(System.currentTimeMillis()) + " - Job accomplished: Product " + product + ", scheduled for " + execution + ", file " + filename);
			pw.close();
		} catch (IOException e) {
			logger.warning("Unable to write to log file " + logFile.getAbsolutePath());
		}
	}
	
	/**
	 * Just removes a file (or folder)
	 * @param path file (or folder) location
	 */
	private void removeFile(String path) {
		new File(path).delete();
	}
	
	/**
	 * Just checks a file (or folder)
	 * @param path file (or folder) location
	 * @return true if file (or folder) exists, false if not
	 */
	private boolean isFileExisting(String path) {
		return new File(path).exists();
	}
	
	/**
	 * generates an extended (without "-") hashcode.
	 * @param eMail the original string
	 * @return encodeed string
	 */
	private String eMailEncode(String eMail) {
		String name = Integer.toString(eMail.hashCode());
		if (name.startsWith("-")) {
			name = "0" + name.substring(1);
		}
		return name;
	}

	/**
	 * Executes the script file that should already be there. This method waits for the script to be finished, so no concurrency(!).
	 * @param scriptFile script location
	 * @return true if execution was successful, false if not
	 */
	private boolean executeScript(String scriptFile) {
		if (!isFileExisting(scriptFile)) {
			logger.warning("Script file " + scriptFile + " does not exist.");
			return false;
		}
		try {
			Runtime r = Runtime.getRuntime();
			Process p = r.exec(scriptFile);
			p.waitFor();
		} catch (IOException e) {
			logger.warning("Unable to execute script file " + scriptFile);
			return false;
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		logger.info("Script file " + scriptFile + " executed.");
		return true;
	}
	
	/**
	 * Just do it.
	 * @param args no args
	 */
	public static void main(String[] args) {
		new Checker().start();
	}
}
